import os
import uuid
import xml.etree.ElementTree as ET

import pytest
import requests

from pypsrp.encryption import WinRMEncryption
from pypsrp.exceptions import (
    AuthenticationError,
    WinRMError,
    WinRMTransportError,
    WSManFaultError,
)
from pypsrp.negotiate import HTTPNegotiateAuth
from pypsrp.wsman import (
    NAMESPACES,
    OptionSet,
    SelectorSet,
    WSMan,
    WSManAction,
    _TransportHTTP,
)

try:
    import requests_credssp
except ImportError:
    requests_credssp = None


class _TransportTest(object):
    def __init__(self, expected_action=None):
        self.endpoint = "testendpoint"
        self.expected_action = expected_action

    def send(self, xml, *args, **kwargs):
        # ensure wsman is always sending a byte string
        assert isinstance(xml, bytes)

        if self.expected_action is None:
            # see what happens if the text is XML but not a WSManFault message
            raise WinRMTransportError("http", 401, "not an XML response")

        req = ET.fromstring(xml)
        action = req.find("s:Header/wsa:Action", NAMESPACES).text
        if action == self.expected_action:
            return """<s:Envelope xml:lang="en-US" xmlns:s="http://www.w3.org/2003/05/soap-envelope" xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/08/addressing">
                <s:Header>
                    <wsa:RelatesTo>uuid:00000000-0000-0000-0000-000000000000</wsa:RelatesTo>
                </s:Header>
                <s:Body>body</s:Body>
            </s:Envelope>"""
        else:
            # we want to set a non XML message as the response text to verify
            # the parsing failure is checked and the original exception is
            # raised
            error_msg = """<s:Envelope xml:lang="en-US" xmlns:s="http://www.w3.org/2003/05/soap-envelope" xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing" xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer" xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing" xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration" xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" xmlns:rsp="http://schemas.microsoft.com/wbem/wsman/1/windows/shell" xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd"><s:Header><a:Action>http://schemas.xmlsoap.org/ws/2004/08/addressing/fault</a:Action><a:MessageID>uuid:4DB571F9-F8DE-48FD-872C-2AF08D996249</a:MessageID><a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To><a:RelatesTo>uuid:eaa98952-3188-458f-b265-b03ace115f20</a:RelatesTo><s:NotUnderstood qname="wsman:ResourceUri" xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" />
                </s:Header>
                <s:Body>
                    <s:Fault>
                        <s:Code>
                            <s:Value>IllegalAction</s:Value>
                        </s:Code>
                        <s:Reason>
                            <s:Text xml:lang="">Illegal action '%s', expecting '%s'</s:Text>
                        </s:Reason>
                    </s:Fault>
                </s:Body>
            </s:Envelope>""" % (
                action,
                self.expected_action,
            )
            raise WinRMTransportError("http", 500, error_msg)


class TestWSMan(object):
    def test_wsman_defaults(self):
        actual = WSMan("")
        assert actual.max_envelope_size == 153600
        assert actual.max_payload_size < actual.max_envelope_size
        assert actual.operation_timeout == 20
        assert isinstance(actual.session_id, str)

        # verify we get a unique session id each time this is initialised
        new_wsman = WSMan("")
        assert actual.session_id != new_wsman.session_id

    def test_override_default(self):
        actual = WSMan("", 8192, 30)
        assert actual.max_envelope_size == 8192
        assert actual.max_payload_size < actual.max_envelope_size
        assert actual.operation_timeout == 30

    def test_invoke_command(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.COMMAND)
        actual = wsman.command("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_connect(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.CONNECT)
        actual = wsman.connect("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_create(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.CREATE)
        actual = wsman.create("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_disconnect(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.DISCONNECT)
        actual = wsman.disconnect("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_enumerate(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.ENUMERATE)
        actual = wsman.enumerate("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_delete(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.DELETE)
        actual = wsman.delete("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_get(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.GET)
        actual = wsman.get("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_pull(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.PULL)
        actual = wsman.pull("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_put(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.PUT)
        actual = wsman.put("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_receive(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.RECEIVE)
        actual = wsman.receive("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_reconnect(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.RECONNECT)
        actual = wsman.reconnect("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_send(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.SEND)
        actual = wsman.send("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_invoke_signal(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.SIGNAL)
        actual = wsman.signal("", None)
        assert actual.tag == "{http://www.w3.org/2003/05/soap-envelope}Body"
        assert actual.text == "body"

    def test_get_header_no_locale(self):
        wsman = WSMan("")
        _, actual = wsman._create_header("action", "resource", None, None, None)
        actual_data_locale = actual.find("wsmv:DataLocale", NAMESPACES)
        actual_locale = actual.find("wsman:Locale", NAMESPACES)

        xml = NAMESPACES["xml"]
        assert actual_data_locale.attrib["{%s}lang" % xml] == "en-US"
        assert actual_locale.attrib["{%s}lang" % xml] == "en-US"

    def test_get_header_explicit_locale(self):
        wsman = WSMan("", locale="en-GB")
        _, actual = wsman._create_header("action", "resource", None, None, None)
        actual_data_locale = actual.find("wsmv:DataLocale", NAMESPACES)
        actual_locale = actual.find("wsman:Locale", NAMESPACES)

        xml = NAMESPACES["xml"]
        assert actual_data_locale.attrib["{%s}lang" % xml] == "en-GB"
        assert actual_locale.attrib["{%s}lang" % xml] == "en-GB"

    def test_get_header_explicit_data_locale(self):
        wsman = WSMan("", data_locale="en-GB")
        _, actual = wsman._create_header("action", "resource", None, None, None)
        actual_data_locale = actual.find("wsmv:DataLocale", NAMESPACES)
        actual_locale = actual.find("wsman:Locale", NAMESPACES)

        xml = NAMESPACES["xml"]
        assert actual_data_locale.attrib["{%s}lang" % xml] == "en-GB"
        assert actual_locale.attrib["{%s}lang" % xml] == "en-US"

    def test_get_header_explicit_both_locale(self):
        wsman = WSMan("", locale="en-AU", data_locale="en-GB")
        _, actual = wsman._create_header("action", "resource", None, None, None)
        actual_data_locale = actual.find("wsmv:DataLocale", NAMESPACES)
        actual_locale = actual.find("wsman:Locale", NAMESPACES)

        xml = NAMESPACES["xml"]
        assert actual_data_locale.attrib["{%s}lang" % xml] == "en-GB"
        assert actual_locale.attrib["{%s}lang" % xml] == "en-AU"

    def test_invoke_mismatch_id(self, monkeypatch):
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000001")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)

        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.SEND)
        with pytest.raises(WinRMError) as exc:
            wsman.send("", None)
        assert (
            str(exc.value) == "Received related id does not match related expected message "
            "id: Sent: uuid:00000000-0000-0000-0000-000000000001, Received: "
            "uuid:00000000-0000-0000-0000-000000000000"
        )

    def test_invoke_transport_error(self):
        wsman = WSMan("")
        wsman.transport = _TransportTest()
        with pytest.raises(WinRMTransportError) as exc:
            wsman.send("", None)
        error_msg = "Bad HTTP response returned from the server. Code: 401, Content: 'not an XML response'"
        assert str(exc.value) == error_msg
        assert exc.value.code == 401
        assert exc.value.protocol == "http"
        assert exc.value.message == error_msg
        assert exc.value.response_text == "not an XML response"

    def test_invoke_wsman_fault(self):
        # we set Create and send Send to cause the test transport to fire the
        # error we want
        wsman = WSMan("")
        wsman.transport = _TransportTest(WSManAction.CREATE)
        with pytest.raises(WSManFaultError) as exc:
            wsman.send("", None)
        error_msg = (
            "Received a WSManFault message. (Code: IllegalAction, Reason: "
            "Illegal action '%s', expecting '%s')" % (WSManAction.SEND, WSManAction.CREATE)
        )
        assert str(exc.value) == error_msg
        assert exc.value.code == "IllegalAction"
        assert exc.value.machine is None
        assert exc.value.message == error_msg
        assert exc.value.provider is None
        assert exc.value.provider_fault is None
        assert exc.value.reason == "Illegal action '%s', expecting '%s'" % (WSManAction.SEND, WSManAction.CREATE)

    def test_raise_native_wsman_fault(self):
        xml_text = """
        <s:Envelope xml:lang="en-US" xmlns:s="http://www.w3.org/2003/05/soap-envelope" xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing" xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer" xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing" xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration" xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" xmlns:rsp="http://schemas.microsoft.com/wbem/wsman/1/windows/shell" xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd"><s:Header><a:Action>http://schemas.xmlsoap.org/ws/2004/08/addressing/fault</a:Action><a:MessageID>uuid:4DB571F9-F8DE-48FD-872C-2AF08D996249</a:MessageID><a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To><a:RelatesTo>uuid:eaa98952-3188-458f-b265-b03ace115f20</a:RelatesTo><s:NotUnderstood qname="wsman:ResourceUri" xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" />
            </s:Header>
            <s:Body>
                <s:Fault>
                    <s:Code>
                        <s:Value>s:MustUnderstand</s:Value>
                    </s:Code>
                    <s:Reason>
                        <s:Text xml:lang="">The WS-Management service cannot process a SOAP header in the request that is marked as mustUnderstand by the client.  This could be caused by the use of a version of the protocol which is not supported, or may be an incompatibility  between the client and server implementations. </s:Text>
                    </s:Reason>
                </s:Fault>
            </s:Body>
        </s:Envelope>"""
        with pytest.raises(WSManFaultError) as exc:
            raise WSMan._parse_wsman_fault(xml_text)
        assert exc.value.code == "s:MustUnderstand"
        assert exc.value.machine is None
        assert (
            exc.value.message == "Received a WSManFault message. (Code: s:MustUnderstand, "
            "Reason: The WS-Management service cannot process a SOAP header "
            "in the request that is marked as mustUnderstand by the client. "
            " This could be caused by the use of a version of the protocol "
            "which is not supported, or may be an incompatibility  between "
            "the client and server implementations.)"
        )
        assert exc.value.provider is None
        assert exc.value.provider_fault is None
        assert exc.value.provider_path is None
        assert (
            exc.value.reason == "The WS-Management service cannot process a SOAP header in the "
            "request that is marked as mustUnderstand by the client.  This "
            "could be caused by the use of a version of the protocol which "
            "is not supported, or may be an incompatibility  between the "
            "client and server implementations."
        )

    def test_raise_native_wsman_fault_no_reason(self):
        xml_text = """
        <s:Envelope xml:lang="en-US" xmlns:s="http://www.w3.org/2003/05/soap-envelope" xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing" xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer" xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing" xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration" xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" xmlns:rsp="http://schemas.microsoft.com/wbem/wsman/1/windows/shell" xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd"><s:Header><a:Action>http://schemas.xmlsoap.org/ws/2004/08/addressing/fault</a:Action><a:MessageID>uuid:4DB571F9-F8DE-48FD-872C-2AF08D996249</a:MessageID><a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To><a:RelatesTo>uuid:eaa98952-3188-458f-b265-b03ace115f20</a:RelatesTo><s:NotUnderstood qname="wsman:ResourceUri" xmlns:wsman="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" />
            </s:Header>
            <s:Body>
                <s:Fault>
                    <s:Code>
                        <s:Value>s:Unknown</s:Value>
                    </s:Code>
                </s:Fault>
            </s:Body>
        </s:Envelope>"""
        with pytest.raises(WSManFaultError) as exc:
            raise WSMan._parse_wsman_fault(xml_text)
        assert exc.value.code == "s:Unknown"
        assert exc.value.machine is None
        assert exc.value.message == "Received a WSManFault message. (Code: s:Unknown)"
        assert exc.value.provider is None
        assert exc.value.provider_fault is None
        assert exc.value.provider_path is None
        assert exc.value.reason is None

    def test_raise_wsman_fault_with_wsman_fault(self):
        xml_text = r"""
        <s:Envelope xml:lang="en-US" xmlns:s="http://www.w3.org/2003/05/soap-envelope" xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing" xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer" xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing" xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration" xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd">
            <s:Header>
                <a:Action>http://schemas.dmtf.org/wbem/wsman/1/wsman/fault</a:Action>
                <a:MessageID>uuid:348D9DCE-B99B-4EBD-A90B-624854B032BB</a:MessageID>
                <a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To>
                <a:RelatesTo>uuid:a82b5f24-7a6c-4170-8cd1-d2031b1203fd</a:RelatesTo>
            </s:Header>
            <s:Body>
                <s:Fault>
                    <s:Code>
                        <s:Value>s:Sender</s:Value>
                        <s:Subcode>
                            <s:Value>w:InvalidParameter</s:Value>
                        </s:Subcode>
                    </s:Code>
                    <s:Reason>
                        <s:Text xml:lang="">The parameter is incorrect. </s:Text>
                    </s:Reason>
                    <s:Detail>
                        <w:FaultDetail>http://schemas.dmtf.org/wbem/wsman/1/wsman/faultDetail/InvalidValue</w:FaultDetail>
                        <f:WSManFault xmlns:f="http://schemas.microsoft.com/wbem/wsman/1/wsmanfault" Code="87" Machine="SERVER2016.domain.local">
                            <f:Message><f:ProviderFault provider="Shell cmd plugin" path="%systemroot%\system32\winrscmd.dll">The parameter is incorrect. </f:ProviderFault></f:Message>
                        </f:WSManFault>
                    </s:Detail>
                </s:Fault>
            </s:Body>
        </s:Envelope>"""
        with pytest.raises(WSManFaultError) as exc:
            raise WSMan._parse_wsman_fault(xml_text)
        assert exc.value.code == 87
        assert exc.value.machine == "SERVER2016.domain.local"
        assert (
            exc.value.message == "Received a WSManFault message. (Code: 87, Machine: "
            "SERVER2016.domain.local, Reason: The parameter is incorrect., "
            "Provider: Shell cmd plugin, Provider Path: %systemroot%\\"
            "system32\\winrscmd.dll, Provider Fault: The parameter is "
            "incorrect.)"
        )
        assert exc.value.provider == "Shell cmd plugin"
        assert exc.value.provider_fault == "The parameter is incorrect."
        assert exc.value.provider_path == "%systemroot%\\system32\\winrscmd.dll"
        assert exc.value.reason == "The parameter is incorrect."

    def test_raise_wsman_fault_without_provider(self):
        xml_text = r"""
        <s:Envelope xml:lang="en-US" xmlns:s="http://www.w3.org/2003/05/soap-envelope" xmlns:a="http://schemas.xmlsoap.org/ws/2004/08/addressing" xmlns:x="http://schemas.xmlsoap.org/ws/2004/09/transfer" xmlns:e="http://schemas.xmlsoap.org/ws/2004/08/eventing" xmlns:n="http://schemas.xmlsoap.org/ws/2004/09/enumeration" xmlns:w="http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd" xmlns:p="http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd">
            <s:Header>
                <a:Action>http://schemas.dmtf.org/wbem/wsman/1/wsman/fault</a:Action>
                <a:MessageID>uuid:EE71C444-1658-4B3F-916D-54CE43B68BC9</a:MessageID>
                <a:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:To>
                <a:RelatesTo>uuid.761ca906-0bf0-41bb-a9d9-4cbbca986aeb</a:RelatesTo>
            </s:Header>
            <s:Body>
                <s:Fault>
                    <s:Code>
                        <s:Value>s:Sender</s:Value>
                        <s:Subcode>
                            <s:Value>w:SchemaValidationError</s:Value>
                        </s:Subcode>
                    </s:Code>
                    <s:Reason>
                        <s:Text xml:lang="">The SOAP XML in the message does not match the corresponding XML schema definition. Change the XML and retry. </s:Text>
                    </s:Reason>
                    <s:Detail>
                        <f:WSManFault xmlns:f="http://schemas.microsoft.com/wbem/wsman/1/wsmanfault" Code="2150858817" Machine="SERVER2008.domain.local">
                            <f:Message>The Windows Remote Shell cannot process the request. The SOAP packet contains an element Argument that is invalid. Retry the request with the correct XML element. </f:Message>
                        </f:WSManFault>
                    </s:Detail>
                </s:Fault>
            </s:Body>
        </s:Envelope>"""
        with pytest.raises(WSManFaultError) as exc:
            raise WSMan._parse_wsman_fault(xml_text)
        assert exc.value.code == 2150858817
        assert exc.value.machine == "SERVER2008.domain.local"
        assert (
            exc.value.message == "Received a WSManFault message. (Code: 2150858817, Machine: "
            "SERVER2008.domain.local, Reason: The Windows Remote Shell "
            "cannot process the request. The SOAP packet contains an "
            "element Argument that is invalid. Retry the request with the "
            "correct XML element.)"
        )
        assert exc.value.provider is None
        assert exc.value.provider_fault is None
        assert exc.value.provider_path is None
        assert (
            exc.value.reason == "The Windows Remote Shell cannot process the request. The SOAP "
            "packet contains an element Argument that is invalid. Retry the "
            "request with the correct XML element."
        )

    def test_wsman_update_envelope_size_explicit(self):
        wsman = WSMan("")
        wsman.update_max_payload_size(4096)
        assert wsman.max_envelope_size == 4096
        # this next value is dependent on a lot of things such as python
        # version and rounding differences, we will just assert against a range
        assert 1450 <= wsman.max_payload_size <= 1835

    @pytest.mark.parametrize(
        "wsman_conn",
        # we just want to validate against different env
        # set on a server
        [[False, "test_wsman_update_envelope_size_150"]],
        indirect=True,
    )
    def test_wsman_update_envelope_size_150(self, wsman_conn):
        wsman_conn.update_max_payload_size()
        assert wsman_conn.max_envelope_size == 153600
        # this next value is dependent on a lot of things such as python
        # version and rounding differences, we will just assert against a range
        assert 113574 <= wsman_conn.max_payload_size <= 113952

    @pytest.mark.parametrize(
        "wsman_conn",
        # we just want to validate against different env
        # set on a server
        [[False, "test_wsman_update_envelope_size_500"]],
        indirect=True,
    )
    def test_wsman_update_envelope_size_500(self, wsman_conn):
        wsman_conn.update_max_payload_size()
        assert wsman_conn.max_envelope_size == 512000
        # this next value is dependent on a lot of things such as python
        # version and rounding differences, we will just assert against a range
        assert 382374 <= wsman_conn.max_payload_size <= 382752

    @pytest.mark.parametrize(
        "wsman_conn",
        # we just want to validate against different env
        # set on a server
        [[False, "test_wsman_update_envelope_size_4096"]],
        indirect=True,
    )
    def test_wsman_update_envelope_size_4096(self, wsman_conn):
        wsman_conn.update_max_payload_size()
        assert wsman_conn.max_envelope_size == 4194304
        # this next value is dependent on a lot of things such as python
        # version and rounding differences, we will just assert against a range
        assert 3144102 <= wsman_conn.max_payload_size <= 3144480


class TestOptionSet(object):
    def test_set_no_options(self):
        option_set = OptionSet()
        actual = option_set.pack()
        assert len(actual.attrib.keys()) == 1
        assert actual.attrib["{http://www.w3.org/2003/05/soap-envelope}mustUnderstand"] == "true"
        assert actual.tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}OptionSet"
        assert actual.text is None
        assert list(actual) == []
        assert str(option_set) == "{}"

    def test_set_one_option(self):
        option_set = OptionSet()
        option_set.add_option("key", "value")
        actual = option_set.pack()
        assert len(actual.attrib.keys()) == 1
        assert actual.attrib["{http://www.w3.org/2003/05/soap-envelope}mustUnderstand"] == "true"
        assert actual.tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}OptionSet"
        assert actual.text is None
        children = list(actual)
        assert len(children) == 1
        assert len(children[0].attrib.keys()) == 1
        assert children[0].attrib["Name"] == "key"
        assert children[0].tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}Option"
        assert children[0].text == "value"
        assert str(option_set) == "{'key': 'value'}"

    def test_set_one_option_with_attributes(self):
        option_set = OptionSet()
        option_set.add_option("key", "value", {"attrib1": "value1", "attrib2": "value2"})
        actual = option_set.pack()
        assert len(actual.attrib.keys()) == 1
        assert actual.attrib["{http://www.w3.org/2003/05/soap-envelope}mustUnderstand"] == "true"
        assert actual.tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}OptionSet"
        assert actual.text is None
        children = list(actual)
        assert len(children) == 1
        assert len(children[0].attrib.keys()) == 3
        assert children[0].attrib["Name"] == "key"
        assert children[0].attrib["attrib1"] == "value1"
        assert children[0].attrib["attrib2"] == "value2"
        assert children[0].tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}Option"
        assert children[0].text == "value"
        assert str(option_set) == "{'key': 'value'}"

    def test_set_multiple_options(self):
        option_set = OptionSet()
        option_set.add_option("key1", "value1")
        option_set.add_option("key2", "value2")
        actual = option_set.pack()
        assert len(actual.attrib.keys()) == 1
        assert actual.attrib["{http://www.w3.org/2003/05/soap-envelope}mustUnderstand"] == "true"
        assert actual.tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}OptionSet"
        assert actual.text is None
        children = list(actual)
        assert len(children) == 2

        assert len(children[0].attrib.keys()) == 1
        assert children[0].attrib["Name"] == "key1"
        assert children[0].tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}Option"
        assert children[0].text == "value1"

        assert len(children[1].attrib.keys()) == 1
        assert children[1].attrib["Name"] == "key2"
        assert children[1].tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}Option"
        assert children[1].text == "value2"

        assert str(option_set) == "{'key1': 'value1', 'key2': 'value2'}"


class TestSelectorSet(object):
    def test_set_no_options(self):
        selector_set = SelectorSet()
        actual = selector_set.pack()
        assert len(actual.attrib.keys()) == 0
        assert actual.tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}SelectorSet"
        assert actual.text is None
        assert list(actual) == []
        assert str(selector_set) == "{}"

    def test_set_one_option(self):
        selector_set = SelectorSet()
        selector_set.add_option("key", "value")
        actual = selector_set.pack()
        assert len(actual.attrib.keys()) == 0
        assert actual.tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}SelectorSet"
        assert actual.text is None
        children = list(actual)
        assert len(children) == 1
        assert len(children[0].attrib.keys()) == 1
        assert children[0].attrib["Name"] == "key"
        assert children[0].tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}Selector"
        assert children[0].text == "value"
        assert str(selector_set) == "{'key': 'value'}"

    def test_set_one_option_with_attributes(self):
        selector_set = SelectorSet()
        selector_set.add_option("key", "value", {"attrib1": "value1", "attrib2": "value2"})
        actual = selector_set.pack()
        assert len(actual.attrib.keys()) == 0
        assert actual.tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}SelectorSet"
        assert actual.text is None
        children = list(actual)
        assert len(children) == 1
        assert len(children[0].attrib.keys()) == 3
        assert children[0].attrib["Name"] == "key"
        assert children[0].attrib["attrib1"] == "value1"
        assert children[0].attrib["attrib2"] == "value2"
        assert children[0].tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}Selector"
        assert children[0].text == "value"
        assert str(selector_set) == "{'key': 'value'}"

    def test_set_multiple_options(self):
        selector_set = SelectorSet()
        selector_set.add_option("key1", "value1")
        selector_set.add_option("key2", "value2")
        actual = selector_set.pack()
        assert len(actual.attrib.keys()) == 0
        assert actual.tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}SelectorSet"
        assert actual.text is None
        children = list(actual)
        assert len(children) == 2

        assert len(children[0].attrib.keys()) == 1
        assert children[0].attrib["Name"] == "key1"
        assert children[0].tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}Selector"
        assert children[0].text == "value1"

        assert len(children[1].attrib.keys()) == 1
        assert children[1].attrib["Name"] == "key2"
        assert children[1].tag == "{http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd}Selector"
        assert children[1].text == "value2"

        assert str(selector_set) == "{'key1': 'value1', 'key2': 'value2'}"


class TestTransportHTTP(object):
    def test_not_supported_auth(self):
        with pytest.raises(ValueError) as err:
            _TransportHTTP("", "", auth="fake")
        assert (
            str(err.value) == "The specified auth 'fake' is not supported, please select one "
            "of 'basic, certificate, credssp, kerberos, negotiate, ntlm'"
        )

    def test_invalid_encryption_value(self):
        with pytest.raises(ValueError) as err:
            _TransportHTTP("", "", encryption="fake")
        assert str(err.value) == "The encryption value 'fake' must be auto, always, or never"

    def test_encryption_always_not_valid_auth_ssl(self):
        with pytest.raises(ValueError) as err:
            _TransportHTTP("", "", auth="basic", encryption="always", ssl=True)
        assert (
            str(err.value) == "Cannot use message encryption with auth 'basic', either set "
            "encryption='auto' or use one of the following auth providers: "
            "credssp, kerberos, negotiate, ntlm"
        )

    def test_encryption_auto_not_valid_auth_no_ssl(self):
        with pytest.raises(ValueError) as err:
            _TransportHTTP("", "", auth="basic", encryption="auto", ssl=False)
        assert (
            str(err.value) == "Cannot use message encryption with auth 'basic', either set "
            "encryption='never', use ssl=True or use one of the following "
            "auth providers: credssp, kerberos, negotiate, ntlm"
        )

    def test_build_basic_no_username(self):
        transport = _TransportHTTP("")
        with pytest.raises(ValueError) as err:
            transport._build_auth_basic(None)
        assert str(err.value) == "For basic auth, the username must be specified"

    def test_build_basic_no_password(self):
        transport = _TransportHTTP("", username="user")
        with pytest.raises(ValueError) as err:
            transport._build_auth_basic(None)
        assert str(err.value) == "For basic auth, the password must be specified"

    def test_build_basic(self):
        transport = _TransportHTTP("", username="user", password="pass", auth="basic")
        session = transport._build_session()
        assert transport.encryption is None
        assert isinstance(session.auth, requests.auth.HTTPBasicAuth)
        assert session.auth.username == "user"
        assert session.auth.password == "pass"

    def test_build_certificate_no_key_pem(self):
        transport = _TransportHTTP("")
        with pytest.raises(ValueError) as err:
            transport._build_auth_certificate(None)
        assert (
            str(err.value) == "For certificate auth, the path to the certificate key pem file "
            "must be specified with certificate_key_pem"
        )

    def test_build_certificate_no_pem(self):
        transport = _TransportHTTP("", certificate_key_pem="path")
        with pytest.raises(ValueError) as err:
            transport._build_auth_certificate(None)
        assert (
            str(err.value) == "For certificate auth, the path to the certificate pem file "
            "must be specified with certificate_pem"
        )

    def test_build_certificate_not_ssl(self):
        transport = _TransportHTTP("", certificate_key_pem="path", certificate_pem="path", ssl=False)
        with pytest.raises(ValueError) as err:
            transport._build_auth_certificate(None)
        assert str(err.value) == "For certificate auth, SSL must be used"

    def test_build_certificate(self):
        transport = _TransportHTTP("", auth="certificate", certificate_key_pem="key_pem", certificate_pem="pem")
        session = transport._build_session()
        assert transport.encryption is None
        assert session.auth is None
        assert session.cert == ("pem", "key_pem")
        assert session.headers["Authorization"] == "http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual"

    @pytest.mark.skipif(
        requests_credssp,
        reason="only raises if requests-credssp is not installed",
    )
    def test_build_credssp_not_imported(self):
        transport = _TransportHTTP("", username="user", password="password")
        with pytest.raises(
            ImportError,
            match=(
                r"Cannot use CredSSP auth as requests-credssp is not "
                r"installed: No module named '?requests_credssp'?"
            ),
        ):
            transport._build_auth_credssp(None)

    def test_build_credssp_no_username(self):
        transport = _TransportHTTP("")
        with pytest.raises(ValueError) as err:
            transport._build_auth_credssp(None)
        assert str(err.value) == "For credssp auth, the username must be specified"

    def test_build_credssp_no_password(self):
        transport = _TransportHTTP("", username="user")
        with pytest.raises(ValueError) as err:
            transport._build_auth_credssp(None)
        assert str(err.value) == "For credssp auth, the password must be specified"

    def test_build_credssp_no_kwargs(self):
        credssp = pytest.importorskip("requests_credssp")

        transport = _TransportHTTP("", username="user", password="pass", auth="credssp")
        session = transport._build_session()
        assert isinstance(session.auth, credssp.HttpCredSSPAuth)
        assert session.auth.disable_tlsv1_2 is False
        assert session.auth.minimum_version == 2
        assert session.auth.password == "pass"
        assert session.auth.username == "user"

    def test_build_credssp_with_kwargs(self):
        credssp = pytest.importorskip("requests_credssp")

        transport = _TransportHTTP(
            "",
            username="user",
            password="pass",
            auth="credssp",
            credssp_auth_mechanism="kerberos",
            credssp_disable_tlsv1_2=True,
            credssp_minimum_version=5,
        )

        session = transport._build_session()
        assert isinstance(session.auth, credssp.HttpCredSSPAuth)
        assert session.auth.disable_tlsv1_2 is True
        assert session.auth.minimum_version == 5
        assert session.auth.password == "pass"
        assert session.auth.username == "user"

    def test_build_kerberos(self):
        transport = _TransportHTTP("", auth="kerberos")
        session = transport._build_session()
        assert isinstance(session.auth, HTTPNegotiateAuth)
        assert session.auth.auth_provider == "kerberos"
        assert session.auth.delegate is False
        assert session.auth.hostname_override is None
        assert session.auth.password is None
        assert session.auth.send_cbt is True
        assert session.auth.service == "host"
        assert session.auth.username is None
        assert session.auth.wrap_required is False

    def test_build_kerberos_with_kwargs(self):
        transport = _TransportHTTP(
            "",
            auth="kerberos",
            username="user",
            ssl=False,
            password="pass",
            negotiate_delegate=True,
            negotiate_hostname_override="host1",
            negotiate_send_cbt=False,
            negotiate_service="HTTP",
        )
        session = transport._build_session()
        assert isinstance(session.auth, HTTPNegotiateAuth)
        assert session.auth.auth_provider == "kerberos"
        assert session.auth.delegate is True
        assert session.auth.hostname_override == "host1"
        assert session.auth.password == "pass"
        assert session.auth.send_cbt is False
        assert session.auth.service == "HTTP"
        assert session.auth.username == "user"
        assert session.auth.wrap_required is True

    def test_build_negotiate(self):
        transport = _TransportHTTP("")
        session = transport._build_session()
        assert isinstance(session.auth, HTTPNegotiateAuth)
        assert session.auth.auth_provider == "negotiate"
        assert session.auth.delegate is False
        assert session.auth.hostname_override is None
        assert session.auth.password is None
        assert session.auth.send_cbt is True
        assert session.auth.service == "host"
        assert session.auth.username is None
        assert session.auth.wrap_required is False

    def test_build_negotiate_with_kwargs(self):
        transport = _TransportHTTP(
            "",
            auth="negotiate",
            username="user",
            ssl=False,
            password="pass",
            negotiate_delegate=True,
            negotiate_hostname_override="host",
            negotiate_send_cbt=False,
            negotiate_service="HTTP",
        )
        session = transport._build_session()
        assert isinstance(session.auth, HTTPNegotiateAuth)
        assert session.auth.auth_provider == "negotiate"
        assert session.auth.delegate is True
        assert session.auth.hostname_override == "host"
        assert session.auth.password == "pass"
        assert session.auth.send_cbt is False
        assert session.auth.service == "HTTP"
        assert session.auth.username == "user"
        assert session.auth.wrap_required is True

    def test_build_ntlm(self):
        transport = _TransportHTTP("", auth="ntlm")
        session = transport._build_session()
        assert isinstance(session.auth, HTTPNegotiateAuth)
        assert session.auth.auth_provider == "ntlm"
        assert session.auth.delegate is False
        assert session.auth.hostname_override is None
        assert session.auth.password is None
        assert session.auth.send_cbt is True
        assert session.auth.service == "host"
        assert session.auth.username is None
        assert session.auth.wrap_required is False

    def test_build_ntlm_with_kwargs(self):
        transport = _TransportHTTP(
            "",
            auth="ntlm",
            username="user",
            ssl=False,
            password="pass",
            negotiate_delegate=True,
            negotiate_hostname_override="host1",
            negotiate_send_cbt=False,
            negotiate_service="HTTP",
            cert_validation=False,
        )
        session = transport._build_session()
        assert isinstance(session.auth, HTTPNegotiateAuth)
        assert session.auth.auth_provider == "ntlm"
        assert session.auth.delegate is True
        assert session.auth.hostname_override == "host1"
        assert session.auth.password == "pass"
        assert session.auth.send_cbt is False
        assert session.auth.service == "HTTP"
        assert session.auth.username == "user"
        assert session.auth.wrap_required is True

    def test_build_session_default(self):
        transport = _TransportHTTP("")
        session = transport._build_session()
        assert session.headers["User-Agent"] == "Python PSRP Client"
        assert session.trust_env is True
        assert isinstance(session.auth, HTTPNegotiateAuth)
        assert "http" not in session.proxies
        assert "https" not in session.proxies
        assert session.verify is True

    def test_build_session_cert_validate(self):
        transport = _TransportHTTP("", cert_validation=True)
        session = transport._build_session()
        assert session.verify is True

    def test_build_session_cert_validate_env(self):
        transport = _TransportHTTP("", cert_validation=True)
        os.environ["REQUESTS_CA_BUNDLE"] = "path_to_REQUESTS_CA_CERT"
        try:
            session = transport._build_session()
        finally:
            del os.environ["REQUESTS_CA_BUNDLE"]
        assert session.verify == "path_to_REQUESTS_CA_CERT"

    def test_build_session_cert_validate_path_override_env(self):
        transport = _TransportHTTP("", cert_validation="kwarg_path")
        os.environ["REQUESTS_CA_BUNDLE"] = "path_to_REQUESTS_CA_CERT"
        try:
            session = transport._build_session()
        finally:
            del os.environ["REQUESTS_CA_BUNDLE"]
        assert session.verify == "kwarg_path"

    def test_build_session_cert_no_validate(self):
        transport = _TransportHTTP("", cert_validation=False)
        session = transport._build_session()
        assert session.verify is False

    def test_build_session_cert_no_validate_override_env(self):
        transport = _TransportHTTP("", cert_validation=False)
        os.environ["REQUESTS_CA_BUNDLE"] = "path_to_REQUESTS_CA_CERT"
        try:
            session = transport._build_session()
        finally:
            del os.environ["REQUESTS_CA_BUNDLE"]
        assert session.verify is False

    def test_build_session_proxies_default(self):
        transport = _TransportHTTP("server")
        session = transport._build_session()
        assert "http" not in session.proxies
        assert "https" not in session.proxies

    def test_build_session_proxies_env(self):
        transport = _TransportHTTP("server")
        os.environ["https_proxy"] = "https://envproxy"
        try:
            session = transport._build_session()
        finally:
            del os.environ["https_proxy"]
        assert "http" not in session.proxies
        assert session.proxies["https"] == "https://envproxy"

    def test_build_session_proxies_kwarg(self):
        transport = _TransportHTTP("server", proxy="https://kwargproxy")
        session = transport._build_session()
        assert "http" not in session.proxies
        assert session.proxies["https"] == "https://kwargproxy"

    def test_build_session_proxies_kwarg_non_ssl(self):
        transport = _TransportHTTP("server", proxy="http://kwargproxy", ssl=False)
        session = transport._build_session()
        assert session.proxies["http"] == "http://kwargproxy"
        assert "https" not in session.proxies

    def test_build_session_proxies_env_kwarg_override(self):
        transport = _TransportHTTP("server", proxy="https://kwargproxy")
        os.environ["https_proxy"] = "https://envproxy"
        try:
            session = transport._build_session()
        finally:
            del os.environ["https_proxy"]
        assert "http" not in session.proxies
        assert session.proxies["https"] == "https://kwargproxy"

    def test_build_session_proxies_env_no_proxy_override(self):
        transport = _TransportHTTP("server", no_proxy=True)
        os.environ["https_proxy"] = "https://envproxy"
        try:
            session = transport._build_session()
        finally:
            del os.environ["https_proxy"]
        assert session.proxies == {"https": False}

    def test_build_session_proxies_kwarg_ignore_no_proxy(self):
        transport = _TransportHTTP("server", proxy="https://kwargproxy", no_proxy=True)
        session = transport._build_session()
        assert "http" not in session.proxies
        assert session.proxies["https"] == "https://kwargproxy"

    def test_send_without_encryption(self, monkeypatch, mocker):
        send_mock = mocker.MagicMock()

        monkeypatch.setattr(_TransportHTTP, "_send_request", send_mock)

        transport = _TransportHTTP("server")
        transport.send(b"message")

        assert send_mock.call_count == 1
        actual_request = send_mock.call_args[0][0]

        assert actual_request.body == b"message"
        assert actual_request.url == "https://server:5986/wsman"
        assert actual_request.headers["content-type"] == "application/soap+xml;charset=UTF-8"

    def test_send_with_encryption(self, monkeypatch, mocker):
        send_mock = mocker.MagicMock()

        def send_request(self, *args, **kwargs):
            self.session.auth.contexts["server"] = mocker.MagicMock()

            return send_mock(*args, **kwargs)

        wrap_mock = mocker.MagicMock()
        wrap_mock.return_value = "multipart/encrypted", b"wrapped"

        monkeypatch.setattr(_TransportHTTP, "_send_request", send_request)
        monkeypatch.setattr(WinRMEncryption, "wrap_message", wrap_mock)

        transport = _TransportHTTP("server", ssl=False)
        transport.send(b"message")
        transport.send(b"message 2")

        assert send_mock.call_count == 3
        actual_request1 = send_mock.call_args_list[0][0][0]
        actual_request2 = send_mock.call_args_list[1][0][0]
        actual_request3 = send_mock.call_args_list[2][0][0]

        assert actual_request1.body is None
        assert actual_request1.url == "http://server:5985/wsman"

        assert actual_request2.body == b"wrapped"
        assert (
            actual_request2.headers["content-type"] == 'multipart/encrypted;protocol="application/'
            'HTTP-SPNEGO-session-encrypted";boundary="Encrypted Boundary"'
        )
        assert actual_request2.url == "http://server:5985/wsman"

        assert actual_request3.body == b"wrapped"
        assert (
            actual_request3.headers["content-type"] == 'multipart/encrypted;protocol="application/'
            'HTTP-SPNEGO-session-encrypted";boundary="Encrypted Boundary"'
        )
        assert actual_request3.url == "http://server:5985/wsman"

        assert wrap_mock.call_count == 2
        assert wrap_mock.call_args_list[0][0][0] == b"message"
        assert wrap_mock.call_args_list[1][0][0] == b"message 2"

    def test_send_default(self, monkeypatch, mocker):
        response = requests.Response()
        response.status_code = 200
        response._content = b"content"
        response.headers["content-type"] = "application/soap+xml;charset=UTF-8"

        send_mock = mocker.MagicMock()
        send_mock.return_value = response

        monkeypatch.setattr(requests.Session, "send", send_mock)

        transport = _TransportHTTP("server", ssl=True)
        session = transport._build_session()
        transport.session = session
        request = requests.Request("POST", transport.endpoint, data=b"data")
        prep_request = session.prepare_request(request)

        actual = transport._send_request(prep_request)
        assert actual == b"content"
        assert send_mock.call_count == 1
        assert send_mock.call_args[0] == (prep_request,)
        assert send_mock.call_args[1]["timeout"] == (30, 30)

    def test_send_timeout_kwargs(self, monkeypatch, mocker):
        response = requests.Response()
        response.status_code = 200
        response._content = b"content"
        response.headers["content-type"] = "application/soap+xml;charset=UTF-8"

        send_mock = mocker.MagicMock()
        send_mock.return_value = response

        monkeypatch.setattr(requests.Session, "send", send_mock)

        transport = _TransportHTTP("server", ssl=True, connection_timeout=20, read_timeout=25)
        session = transport._build_session()
        transport.session = session
        request = requests.Request("POST", transport.endpoint, data=b"data")
        prep_request = session.prepare_request(request)

        actual = transport._send_request(prep_request)
        assert actual == b"content"
        assert send_mock.call_count == 1
        assert send_mock.call_args[0] == (prep_request,)
        assert send_mock.call_args[1]["timeout"] == (20, 25)

    def test_send_auth_error(self, monkeypatch, mocker):
        response = requests.Response()
        response.status_code = 401

        send_mock = mocker.MagicMock()
        send_mock.return_value = response

        monkeypatch.setattr(requests.Session, "send", send_mock)

        transport = _TransportHTTP("server", ssl=True)
        session = transport._build_session()
        transport.session = session
        request = requests.Request("POST", transport.endpoint, data=b"data")
        prep_request = session.prepare_request(request)

        with pytest.raises(AuthenticationError) as err:
            transport._send_request(prep_request)
        assert str(err.value) == "Failed to authenticate the user None with negotiate"

    def test_send_winrm_error_blank(self, monkeypatch, mocker):
        response = requests.Response()
        response.status_code = 500
        response._content = b""

        send_mock = mocker.MagicMock()
        send_mock.return_value = response

        monkeypatch.setattr(requests.Session, "send", send_mock)

        transport = _TransportHTTP("server", ssl=True)
        session = transport._build_session()
        transport.session = session
        request = requests.Request("POST", transport.endpoint, data=b"data")
        prep_request = session.prepare_request(request)

        with pytest.raises(WinRMTransportError) as err:
            transport._send_request(prep_request)
        assert str(err.value) == "Bad HTTP response returned from the server. Code: 500, Content: ''"
        assert err.value.code == 500
        assert err.value.protocol == "http"
        assert err.value.response_text == ""

    def test_send_winrm_error_content(self, monkeypatch, mocker):
        response = requests.Response()
        response.status_code = 500
        response._content = b"error msg"

        send_mock = mocker.MagicMock()
        send_mock.return_value = response

        monkeypatch.setattr(requests.Session, "send", send_mock)

        transport = _TransportHTTP("server", ssl=True)
        session = transport._build_session()
        transport.session = session
        request = requests.Request("POST", transport.endpoint, data=b"data")
        prep_request = session.prepare_request(request)

        with pytest.raises(WinRMTransportError) as err:
            transport._send_request(prep_request)
        assert str(err.value) == "Bad HTTP response returned from the server. Code: 500, Content: 'error msg'"
        assert err.value.code == 500
        assert err.value.protocol == "http"
        assert err.value.response_text == "error msg"

    def test_send_winrm_encrypted_single(self, monkeypatch, mocker):
        response = requests.Response()
        response.status_code = 200
        response._content = b"content"
        response.headers["content-type"] = (
            'multipart/encrypted;protocol="application/HTTP-SPNEGO-session-' 'encrypted";boundary="Encrypted Boundary"'
        )

        send_mock = mocker.MagicMock()
        send_mock.return_value = response
        unwrap_mock = mocker.MagicMock()
        unwrap_mock.return_value = b"unwrapped content"

        monkeypatch.setattr(requests.Session, "send", send_mock)
        monkeypatch.setattr(WinRMEncryption, "unwrap_message", unwrap_mock)

        transport = _TransportHTTP("server", ssl=False)
        transport.encryption = WinRMEncryption(None, None)
        session = transport._build_session()
        transport.session = session
        request = requests.Request("POST", transport.endpoint, data=b"data")
        prep_request = session.prepare_request(request)

        actual = transport._send_request(prep_request)
        assert actual == b"unwrapped content"
        assert send_mock.call_count == 1
        assert send_mock.call_args[0] == (prep_request,)
        assert send_mock.call_args[1]["timeout"] == (30, 30)

        assert unwrap_mock.call_count == 1
        assert unwrap_mock.call_args[0] == (b"content", "Encrypted Boundary")
        assert unwrap_mock.call_args[1] == {}

    def test_send_winrm_encrypted_multiple(self, monkeypatch, mocker):
        response = requests.Response()
        response.status_code = 200
        response._content = b"content"
        response.headers["content-type"] = (
            'multipart/x-multi-encrypted;protocol="application/HTTP-CredSSP-'
            'session-encrypted";boundary="Encrypted Boundary"'
        )

        send_mock = mocker.MagicMock()
        send_mock.return_value = response
        unwrap_mock = mocker.MagicMock()
        unwrap_mock.return_value = b"unwrapped content"

        monkeypatch.setattr(requests.Session, "send", send_mock)
        monkeypatch.setattr(WinRMEncryption, "unwrap_message", unwrap_mock)

        transport = _TransportHTTP("server", ssl=False)
        transport.encryption = WinRMEncryption(None, None)
        session = transport._build_session()
        transport.session = session
        request = requests.Request("POST", transport.endpoint, data=b"data")
        prep_request = session.prepare_request(request)

        actual = transport._send_request(prep_request)
        assert actual == b"unwrapped content"
        assert send_mock.call_count == 1
        assert send_mock.call_args[0] == (prep_request,)
        assert send_mock.call_args[1]["timeout"] == (30, 30)

        assert unwrap_mock.call_count == 1
        assert unwrap_mock.call_args[0] == (b"content", "Encrypted Boundary")
        assert unwrap_mock.call_args[1] == {}

    @pytest.mark.parametrize(
        "ssl, server, port, path, expected",
        [
            [True, "server", 5986, "wsman", "https://server:5986/wsman"],
            [False, "server", 5985, "wsman", "http://server:5985/wsman"],
            [False, "server", 5985, "iis-wsman", "http://server:5985/iis-wsman"],
            [True, "127.0.0.1", 443, "wsman", "https://127.0.0.1:443/wsman"],
            [False, "2001:0db8:0a0b:12f0:0000:0000:0000:0001", 80, "path", "http://[2001:db8:a0b:12f0::1]:80/path"],
            [False, "2001:db8:a0b:12f0::1", 80, "path", "http://[2001:db8:a0b:12f0::1]:80/path"],
            [
                False,
                "2001:0db8:0a0b:12f0:0001:0001:0001:0001",
                5985,
                "wsman",
                "http://[2001:db8:a0b:12f0:1:1:1:1]:5985/wsman",
            ],
            [False, "FE80::0202:B3FF:FE1E:8329", 5985, "wsman", "http://[fe80::202:b3ff:fe1e:8329]:5985/wsman"],
            [
                True,
                "[2001:0db8:0a0b:12f0:0000:0000:0000:0001]",
                5986,
                "wsman",
                "https://[2001:0db8:0a0b:12f0:0000:0000:0000:0001]:5986/wsman",
            ],
        ],
    )
    def test_endpoint_forms(self, ssl, server, port, path, expected):
        actual = _TransportHTTP._create_endpoint(ssl, server, port, path)
        assert actual == expected



import uuid

import pytest
from cryptography.hazmat.primitives.asymmetric import rsa

from pypsrp.complex_objects import (
    Color,
    ControlKeyState,
    Coordinates,
    CultureInfo,
    GenericComplexObject,
    HostMethodIdentifier,
    KeyInfo,
    KeyInfoDotNet,
    ObjectMeta,
    PSCredential,
    Size,
)
from pypsrp.host import PSHost, PSHostRawUserInterface, PSHostUserInterface
from pypsrp.messages import ProgressRecord
from pypsrp.powershell import PowerShell, RunspacePool
from pypsrp.wsman import WSMan


def gen_rsa_keypair(public_exponent, key_size):
    # for a passing test against mocked results, we need the same key
    # for each run

    # private key numbers
    d = 21617696977064831737076881102083866512588021127782856037972563073160418492784722157229048881139551797965254106994729989171004895438848876105403145818526588448619012753150859908505079457128264842711189718538434996465875355972895683414261775421352553376878242391757860294122032733780201193753125763293334069627997366690578378643268496972715558088736335242616446993490560907175346811363459760219154246765368454269083276034064771683561116425318584728938420495690895295417261302673227504350398337950587166908640861984752151206832712112769733009288056773792011545187027621461752928443722277452545442134981026071169761569129
    dmp1 = 129490813503009185789974379255770561425157574567257502592607367628590256916193790244618544180042259447193585640234142495635548617466553819050739804131538251753363069954266652888546939369520788464015077722316584767752481491358983245279072645160828476778216654224017198857901235711822454725809127201111535934753
    dmq1 = 48648800397834409354931458481260042874431001732145738308246185682766008281764239801132482123299506611633669689538582946696791307965087683313609603490735622965494394760749284568774366017328170855872209094830510839953297302962573968041079091285470051311421587284299491265676844471173504726139132522742725727613
    iqmp = 39631089624266744377721024140775526581242717587318543319284349361749959043577498327792652579346928867008103676271384098405811933888254575054903735904994321302658687548878487654303384553639708412577672820397125886631056454780398875240593971733447420281054575014889376749144359968455995345232769795875325417317
    p = 174331130742537243408330955079815843963035167715989214203198624738597363753422194561659111132445142920926993058833709875440980890363712769908141629896643845627776407469038077862709547359496382776626050712560846670587394813048350142683947903572416102765283173128172045224943986952459084128294017448217886823253
    q = 141639112913055250372907285405879139487409354087944702421480687298597773645578986399236760591500655922107321654521995138945695777437048045372013886449237695348272152351631535441304559114954968164429007883070066144736528002854651716726263900736377199146896337887655964419309789253989248300176824884184363216819

    # public key numbers
    e = 65537
    n = 24692106711502830011203227021058444318027801046612842012663747949974978593541529463344548800446917862266219189049856550539417324579114258210080798360122994007305091566363663241781504651372764226027210216355916383975880112316706422502404691353489765771310270171473497918954906308690817673196552680498374521519566949221302301125182104198985782657283395055909134373469597836420671163965867038455758131344733786842283328454828820406016508955409107145350345035248825315933976893356673777910251028486191789752627573225093968284278302684745743589192378470115772764709506475265246795419324395050366115533203601201395969892207

    public_numbers = rsa.RSAPublicNumbers(e, n)
    numbers = rsa.RSAPrivateNumbers(p, q, d, dmp1, dmq1, iqmp, public_numbers)
    key = numbers.private_key()

    return key


class TestPSHost(object):
    @pytest.mark.parametrize(
        "wsman_conn",
        # The actual culture can vary from host to host
        [[True, "test_psrp_pshost_methods"]],
        indirect=True,
    )
    def test_psrp_pshost_methods(self, wsman_conn):
        host = PSHost(None, None, False, "host name", None, None, "1.0")

        with RunspacePool(wsman_conn, host=host) as pool:
            ps = PowerShell(pool)
            # SetShouldExit is really the only one that seems to work so
            # we will just test that
            ps.add_script("$host.CurrentCulture; $host.SetShouldExit(1)")
            actual = ps.invoke()
            assert len(actual) == 1
            assert str(actual[0]) == "en-US"
            assert isinstance(actual[0], CultureInfo)
            assert host.rc == 1

    def test_pshost_not_implemented_methods(self):
        host = PSHost(None, None, False, None, None, None, None)
        with pytest.raises(NotImplementedError):
            host.EnterNestedPrompt(None, None)
        with pytest.raises(NotImplementedError):
            host.ExitNestedPrompt(None, None)

    def test_pshost_methods(self):
        wsman = WSMan("server")
        runspace = RunspacePool(wsman)
        host = PSHost(CultureInfo(), CultureInfo(), True, "name", None, None, "1.0")

        assert host.GetName(None, None) == "name"
        actual_version = host.GetVersion(runspace, None)
        assert actual_version.text == "1.0"
        assert actual_version.tag == "Version"
        assert isinstance(host.GetInstanceId(None, None), uuid.UUID)
        assert isinstance(host.GetCurrentCulture(None, None), CultureInfo)
        assert isinstance(host.GetCurrentUICulture(None, None), CultureInfo)
        host.NotifyBeginApplication(None, None)
        host.NotifyEndApplication(None, None)

    def test_pshost_run_method_not_implemented(self):
        host = PSHost(None, None, False, None, None, None, None)
        actual = host.run_method(HostMethodIdentifier(value=53), [], None)
        assert actual is None


class TestPSHostUserInterface(object):
    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_pshost_ui_mocked_methods"]], indirect=True)
    def test_psrp_pshost_ui_mocked_methods(self, wsman_conn, monkeypatch, mocker):
        # This tests that the args from an actual host call match up with our
        # definitions
        monkeypatch.setattr("cryptography.hazmat.primitives.asymmetric.rsa.generate_private_key", gen_rsa_keypair)

        mock_read_line = mocker.MagicMock(return_value="ReadLine response")
        mock_read_line_as_ss = mocker.MagicMock()
        mock_write1 = mocker.MagicMock(return_value=None)
        mock_write2 = mocker.MagicMock(return_value=None)
        mock_write_line1 = mocker.MagicMock(return_value=None)
        mock_write_line = mocker.MagicMock(return_value=None)
        mock_write_line3 = mocker.MagicMock(return_value=None)
        mock_write_error = mocker.MagicMock(return_value=None)
        mock_write_debug = mocker.MagicMock(return_value=None)
        mock_write_progress = mocker.MagicMock(return_value=None)
        mock_write_verbose = mocker.MagicMock(return_value=None)
        mock_write_warning = mocker.MagicMock(return_value=None)
        mock_prompt = mocker.MagicMock(
            return_value={
                "prompt field": "prompt response",
            }
        )
        mock_prompt_credential = mocker.MagicMock()
        mock_prompt_choice = mocker.MagicMock(return_value=1)

        host_ui = PSHostUserInterface()
        host_ui.ReadLine = mock_read_line
        host_ui.ReadLineAsSecureString = mock_read_line_as_ss
        host_ui.Write1 = mock_write1
        host_ui.Write2 = mock_write2
        host_ui.WriteLine1 = mock_write_line1
        host_ui.WriteLine2 = mock_write_line
        host_ui.WriteLine3 = mock_write_line3
        host_ui.WriteErrorLine = mock_write_error
        host_ui.WriteDebugLine = mock_write_debug
        host_ui.WriteProgress = mock_write_progress
        host_ui.WriteVerboseLine = mock_write_verbose
        host_ui.WriteWarningLine = mock_write_warning
        host_ui.Prompt = mock_prompt
        # seems like PS never calls PromptForCredential1 so we will skip that
        host_ui.PromptForCredential2 = mock_prompt_credential
        host_ui.PromptForChoice = mock_prompt_choice

        host = PSHost(None, None, False, None, None, host_ui, None)

        with RunspacePool(wsman_conn, host=host) as pool:
            pool.exchange_keys()
            mock_read_line_as_ss.return_value = pool.serialize("ReadLineAsSecureString response", ObjectMeta("SS"))
            mock_ps_credential = PSCredential(username="username", password="password")
            mock_prompt_credential.return_value = mock_ps_credential

            ps = PowerShell(pool)
            ps.add_script(
                """$host.UI.ReadLine()
$host.UI.ReadLineAsSecureString()
$host.UI.Write("Write1")
$host.UI.Write([System.ConsoleColor]::Blue, [System.ConsoleColor]::White, "Write2")
$host.UI.WriteLine()
$host.UI.WriteLine("WriteLine2")
$host.UI.WriteLine([System.ConsoleColor]::Gray, [System.ConsoleColor]::Green, "WriteLine3")
$host.UI.WriteErrorLine("WriteErrorLine")
$host.UI.WriteDebugLine("WriteDebugLine")
$host.UI.WriteProgress(1, (New-Object -TypeName System.Management.Automation.ProgressRecord -ArgumentList 2, "activity", "description"))
$host.UI.WriteVerboseLine("WriteVerboseLine")
$host.UI.WriteWarningLine("WriteWarningLine")

$prompt_field = New-Object -TypeName System.Management.Automation.Host.FieldDescription -ArgumentList @("prompt field")
$prompt_field.Label = "PromptLabel"
$host.UI.Prompt("Prompt caption", "Prompt message", $prompt_field)

$host.UI.PromptForCredential("PromptForCredential caption", "PromptForCredential message", "PromptForCredential user", "PromptForCredential target")

$choice_field1 = New-Object -TypeName System.Management.Automation.Host.ChoiceDescription -ArgumentList "Prompt1 label", "Prompt1 help"
$choice_field2 = New-Object -TypeName System.Management.Automation.Host.ChoiceDescription -ArgumentList "Prompt2 label", "Prompt2 help"
$host.UI.PromptForChoice("PromptForChoice caption", "PromptForChoice message", @($choice_field1, $choice_field2), 0)"""
            )
            actual = ps.invoke()

        assert len(actual) == 5

        assert actual[0] == "ReadLine response"
        assert mock_read_line.call_count == 1
        assert isinstance(mock_read_line.call_args[0][0], RunspacePool)
        assert isinstance(mock_read_line.call_args[0][1], PowerShell)

        assert actual[1] == "ReadLineAsSecureString response"
        assert mock_read_line_as_ss.call_count == 1
        assert isinstance(mock_read_line_as_ss.call_args[0][0], RunspacePool)
        assert isinstance(mock_read_line_as_ss.call_args[0][1], PowerShell)

        assert mock_write1.call_count == 1
        assert isinstance(mock_write1.call_args[0][0], RunspacePool)
        assert isinstance(mock_write1.call_args[0][1], PowerShell)
        assert mock_write1.call_args[0][2] == "Write1"

        assert mock_write2.call_count == 1
        assert isinstance(mock_write2.call_args[0][0], RunspacePool)
        assert isinstance(mock_write2.call_args[0][1], PowerShell)
        assert mock_write2.call_args[0][2] == Color.BLUE
        assert mock_write2.call_args[0][3] == Color.WHITE
        assert mock_write2.call_args[0][4] == "Write2"

        assert mock_write_line1.call_count == 1
        assert isinstance(mock_write_line1.call_args[0][0], RunspacePool)
        assert isinstance(mock_write_line1.call_args[0][1], PowerShell)

        assert mock_write_line.call_count == 1
        assert isinstance(mock_write_line.call_args[0][0], RunspacePool)
        assert isinstance(mock_write_line.call_args[0][1], PowerShell)
        assert mock_write_line.call_args[0][2] == "WriteLine2"

        assert mock_write_line3.call_count == 1
        assert isinstance(mock_write_line3.call_args[0][0], RunspacePool)
        assert isinstance(mock_write_line3.call_args[0][1], PowerShell)
        assert mock_write_line3.call_args[0][2] == Color.GRAY
        assert mock_write_line3.call_args[0][3] == Color.GREEN
        assert mock_write_line3.call_args[0][4] == "WriteLine3"

        assert mock_write_error.call_count == 1
        assert isinstance(mock_write_error.call_args[0][0], RunspacePool)
        assert isinstance(mock_write_error.call_args[0][1], PowerShell)
        assert mock_write_error.call_args[0][2] == "WriteErrorLine"

        assert mock_write_debug.call_count == 1
        assert isinstance(mock_write_debug.call_args[0][0], RunspacePool)
        assert isinstance(mock_write_debug.call_args[0][1], PowerShell)
        assert mock_write_debug.call_args[0][2] == "WriteDebugLine"

        # On PSv5 a progress record is always sent, we still sent one
        # ourselves to ensure it works so we verify we received at least
        # one and assert the last
        assert mock_write_progress.call_count > 0
        progress_args = mock_write_progress.call_args_list[-1]
        assert isinstance(progress_args[0][0], RunspacePool)
        assert isinstance(progress_args[0][1], PowerShell)
        assert progress_args[0][2] == 1
        progress_record = pool._serializer.deserialize(progress_args[0][3], ObjectMeta("Obj", object=ProgressRecord))
        assert progress_record.activity == "activity"
        assert progress_record.activity_id == 2
        assert progress_record.description == "description"

        assert mock_write_verbose.call_count == 1
        assert isinstance(mock_write_verbose.call_args[0][0], RunspacePool)
        assert isinstance(mock_write_verbose.call_args[0][1], PowerShell)
        assert mock_write_verbose.call_args[0][2] == "WriteVerboseLine"

        assert mock_write_warning.call_count == 1
        assert isinstance(mock_write_warning.call_args[0][0], RunspacePool)
        assert isinstance(mock_write_warning.call_args[0][1], PowerShell)
        assert mock_write_warning.call_args[0][2] == "WriteWarningLine"

        assert actual[2] == {"prompt field": "prompt response"}
        assert mock_prompt.call_count == 1
        assert isinstance(mock_prompt.call_args[0][0], RunspacePool)
        assert isinstance(mock_prompt.call_args[0][1], PowerShell)
        assert mock_prompt.call_args[0][2] == "Prompt caption"
        assert mock_prompt.call_args[0][3] == "Prompt message"
        assert isinstance(mock_prompt.call_args[0][4], list)
        assert len(mock_prompt.call_args[0][4]) == 1
        assert mock_prompt.call_args[0][4][0].extended_properties["name"] == "prompt field"
        assert mock_prompt.call_args[0][4][0].extended_properties["label"] == "PromptLabel"

        assert isinstance(actual[3], PSCredential)
        assert actual[3].username == "username"
        assert actual[3].password == "password"
        assert mock_prompt_credential.call_count == 1
        assert isinstance(mock_prompt_credential.call_args[0][0], RunspacePool)
        assert isinstance(mock_prompt_credential.call_args[0][1], PowerShell)
        assert mock_prompt_credential.call_args[0][2] == "PromptForCredential caption"
        assert mock_prompt_credential.call_args[0][3] == "PromptForCredential message"
        assert mock_prompt_credential.call_args[0][4] == "PromptForCredential user"
        assert mock_prompt_credential.call_args[0][5] == "PromptForCredential target"
        assert mock_prompt_credential.call_args[0][6] == 3
        assert mock_prompt_credential.call_args[0][7] == 1

        assert actual[4] == 1
        assert mock_prompt_choice.call_count == 1
        assert isinstance(mock_prompt_choice.call_args[0][0], RunspacePool)
        assert isinstance(mock_prompt_choice.call_args[0][1], PowerShell)
        assert mock_prompt_choice.call_args[0][2] == "PromptForChoice caption"
        assert mock_prompt_choice.call_args[0][3] == "PromptForChoice message"
        assert isinstance(mock_prompt_choice.call_args[0][4], list)
        assert len(mock_prompt_choice.call_args[0][4]) == 2
        assert mock_prompt_choice.call_args[0][4][0].extended_properties["label"] == "Prompt1 label"
        assert mock_prompt_choice.call_args[0][4][0].extended_properties["helpMessage"] == "Prompt1 help"
        assert mock_prompt_choice.call_args[0][4][1].extended_properties["label"] == "Prompt2 label"
        assert mock_prompt_choice.call_args[0][4][1].extended_properties["helpMessage"] == "Prompt2 help"

    def test_ps_host_ui_implemented(self):
        ui = PSHostUserInterface()
        ui.Write1(None, None, "value")
        assert ui.stdout == ["value"]

        ui = PSHostUserInterface()
        ui.Write2(None, None, 1, 1, "value")
        assert ui.stdout == ["value"]

        ui = PSHostUserInterface()
        ui.WriteLine1(None, None)
        assert ui.stdout == ["\r\n"]

        ui = PSHostUserInterface()
        ui.WriteLine2(None, None, "value")
        assert ui.stdout == ["value\r\n"]

        ui = PSHostUserInterface()
        ui.WriteLine3(None, None, 1, 1, "value")
        assert ui.stdout == ["value\r\n"]

        ui = PSHostUserInterface()
        ui.WriteErrorLine(None, None, "value")
        assert ui.stderr == ["value\r\n"]

        ui = PSHostUserInterface()
        ui.WriteDebugLine(None, None, "value")
        assert ui.stdout == ["DEBUG: value\r\n"]

        ui = PSHostUserInterface()
        ui.WriteProgress(None, None, 1, None)
        assert ui.stdout == []

        ui = PSHostUserInterface()
        ui.WriteVerboseLine(None, None, "value")
        assert ui.stdout == ["VERBOSE: value\r\n"]

        ui = PSHostUserInterface()
        ui.WriteWarningLine(None, None, "value")
        assert ui.stdout == ["WARNING: value\r\n"]

    def test_pshost_ui_not_implemented(self):
        # tests that methods we have marked as not implemented raise the error
        ui = PSHostUserInterface()

        with pytest.raises(NotImplementedError):
            ui.ReadLine(None, None)

        with pytest.raises(NotImplementedError):
            ui.ReadLineAsSecureString(None, None)

        with pytest.raises(NotImplementedError):
            ui.Prompt(None, None, None, None, None)

        with pytest.raises(NotImplementedError):
            ui.PromptForCredential1(None, None, None, None, None, None)

        with pytest.raises(NotImplementedError):
            ui.PromptForCredential2(None, None, None, None, None, None, None, None)

        with pytest.raises(NotImplementedError):
            ui.PromptForChoice(None, None, None, None, None, None)


class TestPSHostRawUserInterface(object):
    @pytest.mark.parametrize("wsman_conn", [[True, "test_psrp_pshost_raw_ui_mocked_methods"]], indirect=True)
    def test_psrp_pshost_raw_ui_mocked_methods(self, wsman_conn, monkeypatch, mocker):
        # in a mocked context the calculated size differs on a few variables
        # we will mock out that call and return the ones used in our existing
        # responses
        mock_calc = mocker.MagicMock()
        mock_calc.side_effect = [113955, 382750]

        key_info = KeyInfo(code=65, character="a", state=ControlKeyState.CapsLockOn, key_down=True)

        set_foreground_color = mocker.MagicMock(return_value=None)
        set_background_color = mocker.MagicMock(return_value=None)
        set_cursor_position = mocker.MagicMock(return_value=None)
        set_window_position = mocker.MagicMock(return_value=None)
        set_cursor_size = mocker.MagicMock(return_value=None)
        set_buffer_size = mocker.MagicMock(return_value=None)
        set_window_size = mocker.MagicMock(return_value=None)
        set_window_title = mocker.MagicMock(return_value=None)
        read_key = mocker.MagicMock(return_value=key_info)
        flush_input = mocker.MagicMock(return_value=None)
        set_buffer1 = mocker.MagicMock(return_value=None)
        set_buffer2 = mocker.MagicMock(return_value=None)
        scroll_buffer = mocker.MagicMock(return_value=None)

        window_title = "pypsrp window"
        cursor_size = 50
        foreground_color = Color(value=Color.WHITE)
        background_color = Color(value=Color.BLUE)
        cursor_position = Coordinates(x=1, y=2)
        window_position = Coordinates(x=3, y=4)
        buffer_size = Size(width=80, height=80)
        max_physical_window_size = Size(width=80, height=80)
        max_window_size = Size(width=80, height=80)
        window_size = Size(width=80, height=80)

        host_raw_ui = PSHostRawUserInterface(
            window_title,
            cursor_size,
            foreground_color,
            background_color,
            cursor_position,
            window_position,
            buffer_size,
            max_physical_window_size,
            max_window_size,
            window_size,
        )
        host_raw_ui.SetForegroundColor = set_foreground_color
        host_raw_ui.SetBackgroundColor = set_background_color
        host_raw_ui.SetCursorPosition = set_cursor_position
        host_raw_ui.SetWindowPosition = set_window_position
        host_raw_ui.SetCursorSize = set_cursor_size
        host_raw_ui.SetBufferSize = set_buffer_size
        host_raw_ui.SetWindowSize = set_window_size
        host_raw_ui.SetWindowTitle = set_window_title
        host_raw_ui.ReadKey = read_key
        host_raw_ui.FlushInputBuffer = flush_input
        host_raw_ui.SetBufferContents1 = set_buffer1
        host_raw_ui.SetBufferContents2 = set_buffer2
        host_raw_ui.ScrollBufferContents = scroll_buffer

        host_ui = PSHostUserInterface(host_raw_ui)
        host = PSHost(None, None, False, None, None, host_ui, None)

        with RunspacePool(wsman_conn, host=host) as pool:
            ps = PowerShell(pool)
            ps.add_script(
                """$host.UI.RawUI.ForegroundColor
$host.UI.RawUI.ForegroundColor = [System.ConsoleColor]::Green
$host.UI.RawUI.ForegroundColor

$host.UI.RawUI.BackgroundColor
$host.UI.RawUI.BackgroundColor = [System.ConsoleColor]::Red
$host.UI.RawUI.BackgroundColor

$host.UI.RawUI.CursorPosition
$host.UI.RawUI.CursorPosition = (New-Object -TypeName System.Management.Automation.Host.Coordinates -ArgumentList 11, 12)
$host.UI.RawUI.CursorPosition

$host.UI.RawUI.WindowPosition
$host.UI.RawUI.WindowPosition = (New-Object -TypeName System.Management.Automation.Host.Coordinates -ArgumentList 13, 14)
$host.UI.RawUI.WindowPosition

$host.UI.RawUI.CursorSize
$host.UI.RawUI.CursorSize = 25
$host.UI.RawUI.CursorSize

$host.UI.RawUI.BufferSize
$host.UI.RawUI.BufferSize = (New-Object -TypeName System.Management.Automation.Host.Size -ArgumentList 8, 9)
$host.UI.RawUI.BufferSize

$host.UI.RawUI.WindowSize
$host.UI.RawUI.WindowSize = (New-Object -TypeName System.Management.Automation.Host.Size -ArgumentList 8, 9)
$host.UI.RawUI.WindowSize

$host.UI.RawUI.WindowTitle
$host.UI.RawUI.WindowTitle = "New Window Title"
$host.UI.RawUI.WindowTitle

$host.UI.RawUI.ReadKey()

$host.UI.RawUI.FlushInputBuffer()

$buffer_cell = New-Object -TypeName System.Management.Automation.Host.BufferCell -ArgumentList "Z", ([System.ConsoleColor]::Red), ([System.ConsoleColor]::Green), ([System.Management.Automation.Host.BufferCellType]::Complete)
$rectangle = New-Object -TypeName System.Management.Automation.Host.Rectangle -ArgumentList 1, 2, 3, 4
$host.UI.RawUI.SetBufferContents($rectangle, $buffer_cell)

$coordinates = New-Object -TypeName System.Management.Automation.Host.Coordinates -ArgumentList 15, 15

$buffer_cell1_1 = New-Object -TypeName System.Management.Automation.Host.BufferCell -ArgumentList "A", ([System.ConsoleColor]::Black), ([System.ConsoleColor]::White), ([System.Management.Automation.Host.BufferCellType]::Leading)
$buffer_cell1_2 = New-Object -TypeName System.Management.Automation.Host.BufferCell -ArgumentList "B", ([System.ConsoleColor]::Black), ([System.ConsoleColor]::White), ([System.Management.Automation.Host.BufferCellType]::Trailing)
$buffer_cell2_1 = New-Object -TypeName System.Management.Automation.Host.BufferCell -ArgumentList "C", ([System.ConsoleColor]::Black), ([System.ConsoleColor]::White), ([System.Management.Automation.Host.BufferCellType]::Leading)
$buffer_cell2_2 = New-Object -TypeName System.Management.Automation.Host.BufferCell -ArgumentList "D", ([System.ConsoleColor]::Black), ([System.ConsoleColor]::White), ([System.Management.Automation.Host.BufferCellType]::Trailing)

$cells = New-Object -TypeName "System.Management.Automation.Host.BufferCell[,]" -ArgumentList 2,2
$cells[0,0] = $buffer_cell1_1
$cells[0,1] = $buffer_cell1_2
$cells[1,1] = $buffer_cell2_2
$cells[1,0] = $buffer_cell2_1
$host.UI.RawUI.SetBufferContents($coordinates, $cells)

$host.UI.RawUI.ScrollBufferContents($rectangle, $coordinates, $rectangle, $buffer_cell)"""
            )
            actual = ps.invoke()

        assert len(actual) == 17

        assert str(actual[0]) == "White"
        assert str(actual[1]) == "Green"
        assert set_foreground_color.call_count == 1
        assert isinstance(set_foreground_color.call_args[0][0], RunspacePool)
        assert isinstance(set_foreground_color.call_args[0][1], PowerShell)
        assert set_foreground_color.call_args[0][2] == Color.GREEN

        assert str(actual[2]) == "Blue"
        assert str(actual[3]) == "Red"
        assert set_background_color.call_count == 1
        assert isinstance(set_background_color.call_args[0][0], RunspacePool)
        assert isinstance(set_background_color.call_args[0][1], PowerShell)
        assert set_background_color.call_args[0][2] == Color.RED

        assert str(actual[4]) == "1,2"
        assert str(actual[5]) == "11,12"
        assert set_cursor_position.call_count == 1
        assert isinstance(set_cursor_position.call_args[0][0], RunspacePool)
        assert isinstance(set_cursor_position.call_args[0][1], PowerShell)
        assert set_cursor_position.call_args[0][2].extended_properties["x"] == 11
        assert set_cursor_position.call_args[0][2].extended_properties["y"] == 12

        assert str(actual[6]) == "3,4"
        assert str(actual[7]) == "13,14"
        assert set_window_position.call_count == 1
        assert isinstance(set_window_position.call_args[0][0], RunspacePool)
        assert isinstance(set_window_position.call_args[0][1], PowerShell)
        assert set_window_position.call_args[0][2].extended_properties["x"] == 13
        assert set_window_position.call_args[0][2].extended_properties["y"] == 14

        assert actual[8] == 50
        assert actual[9] == 25
        assert set_cursor_size.call_count == 1
        assert isinstance(set_cursor_size.call_args[0][0], RunspacePool)
        assert isinstance(set_cursor_size.call_args[0][1], PowerShell)
        assert set_cursor_size.call_args[0][2] == 25

        assert str(actual[10]) == "80,80"
        assert str(actual[11]) == "8,9"
        assert set_buffer_size.call_count == 1
        assert isinstance(set_buffer_size.call_args[0][0], RunspacePool)
        assert isinstance(set_buffer_size.call_args[0][1], PowerShell)
        assert isinstance(set_buffer_size.call_args[0][2], GenericComplexObject)
        assert set_buffer_size.call_args[0][2].extended_properties["width"] == 8
        assert set_buffer_size.call_args[0][2].extended_properties["height"] == 9

        assert str(actual[12]) == "80,80"
        assert str(actual[13]) == "8,9"
        assert set_window_size.call_count == 1
        assert isinstance(set_window_size.call_args[0][0], RunspacePool)
        assert isinstance(set_window_size.call_args[0][1], PowerShell)
        assert isinstance(set_window_size.call_args[0][2], GenericComplexObject)
        assert set_window_size.call_args[0][2].extended_properties["width"] == 8
        assert set_window_size.call_args[0][2].extended_properties["height"] == 9

        assert actual[14] == "pypsrp window"
        assert actual[15] == "New Window Title"
        assert set_window_title.call_count == 1
        assert isinstance(set_window_title.call_args[0][0], RunspacePool)
        assert isinstance(set_window_title.call_args[0][1], PowerShell)
        assert set_window_title.call_args[0][2] == "New Window Title"

        assert str(actual[16]) == "65,a,CapsLockOn,True"
        assert isinstance(actual[16], KeyInfoDotNet)
        assert read_key.call_count == 1
        assert isinstance(read_key.call_args[0][0], RunspacePool)
        assert isinstance(read_key.call_args[0][1], PowerShell)
        assert read_key.call_args[0][2] == 4

        assert flush_input.call_count == 1
        assert isinstance(flush_input.call_args[0][0], RunspacePool)
        assert isinstance(flush_input.call_args[0][1], PowerShell)

        assert set_buffer1.call_count == 1
        assert isinstance(set_buffer1.call_args[0][0], RunspacePool)
        assert isinstance(set_buffer1.call_args[0][1], PowerShell)
        assert isinstance(set_buffer1.call_args[0][2], GenericComplexObject)
        assert set_buffer1.call_args[0][2].extended_properties["left"] == 1
        assert set_buffer1.call_args[0][2].extended_properties["top"] == 2
        assert set_buffer1.call_args[0][2].extended_properties["right"] == 3
        assert set_buffer1.call_args[0][2].extended_properties["bottom"] == 4
        fill = set_buffer1.call_args[0][3]
        assert isinstance(fill, GenericComplexObject)
        assert fill.extended_properties["character"] == "Z"
        assert fill.extended_properties["foregroundColor"] == 12
        assert fill.extended_properties["backgroundColor"] == 10
        assert fill.extended_properties["bufferCellType"] == 0

        assert set_buffer2.call_count == 1
        assert isinstance(set_buffer2.call_args[0][0], RunspacePool)
        assert isinstance(set_buffer2.call_args[0][1], PowerShell)
        assert isinstance(set_buffer2.call_args[0][2], GenericComplexObject)
        assert set_buffer2.call_args[0][2].extended_properties["x"] == 15
        assert set_buffer2.call_args[0][2].extended_properties["y"] == 15
        assert isinstance(set_buffer2.call_args[0][3], GenericComplexObject)
        assert set_buffer2.call_args[0][3].extended_properties["mal"] == [2, 2]
        set_contents = set_buffer2.call_args[0][3].extended_properties["mae"]
        assert len(set_contents) == 4
        assert set_contents[0].extended_properties["character"] == "A"
        assert set_contents[0].extended_properties["foregroundColor"] == 0
        assert set_contents[0].extended_properties["backgroundColor"] == 15
        assert set_contents[0].extended_properties["bufferCellType"] == 1
        assert set_contents[1].extended_properties["character"] == "B"
        assert set_contents[1].extended_properties["foregroundColor"] == 0
        assert set_contents[1].extended_properties["backgroundColor"] == 15
        assert set_contents[1].extended_properties["bufferCellType"] == 2
        assert set_contents[2].extended_properties["character"] == "C"
        assert set_contents[2].extended_properties["foregroundColor"] == 0
        assert set_contents[2].extended_properties["backgroundColor"] == 15
        assert set_contents[2].extended_properties["bufferCellType"] == 1
        assert set_contents[3].extended_properties["character"] == "D"
        assert set_contents[3].extended_properties["foregroundColor"] == 0
        assert set_contents[3].extended_properties["backgroundColor"] == 15
        assert set_contents[3].extended_properties["bufferCellType"] == 2

        assert scroll_buffer.call_count == 1
        assert isinstance(scroll_buffer.call_args[0][0], RunspacePool)
        assert isinstance(scroll_buffer.call_args[0][1], PowerShell)
        source = scroll_buffer.call_args[0][2]
        assert isinstance(source, GenericComplexObject)
        assert source.extended_properties["left"] == 1
        assert source.extended_properties["top"] == 2
        assert source.extended_properties["right"] == 3
        assert source.extended_properties["bottom"] == 4
        destination = scroll_buffer.call_args[0][3]
        assert isinstance(destination, GenericComplexObject)
        assert destination.extended_properties["x"] == 15
        assert destination.extended_properties["y"] == 15
        clip = scroll_buffer.call_args[0][4]
        assert isinstance(clip, GenericComplexObject)
        assert clip.extended_properties["left"] == 1
        assert clip.extended_properties["top"] == 2
        assert clip.extended_properties["right"] == 3
        assert clip.extended_properties["bottom"] == 4
        fill = scroll_buffer.call_args[0][5]
        assert isinstance(fill, GenericComplexObject)
        assert fill.extended_properties["character"] == "Z"
        assert fill.extended_properties["foregroundColor"] == 12
        assert fill.extended_properties["backgroundColor"] == 10
        assert fill.extended_properties["bufferCellType"] == 0

    def test_ps_host_raw_ui_method(self):
        window_title = "pypsrp window"
        cursor_size = 50
        foreground_color = Color(value=Color.WHITE)
        background_color = Color(value=Color.BLUE)
        cursor_position = Coordinates(x=1, y=2)
        window_position = Coordinates(x=3, y=4)
        buffer_size = Size(width=80, height=80)
        max_physical_window_size = Size(width=80, height=80)
        max_window_size = Size(width=80, height=80)
        window_size = Size(width=80, height=80)

        raw_ui = PSHostRawUserInterface(
            window_title,
            cursor_size,
            foreground_color,
            background_color,
            cursor_position,
            window_position,
            buffer_size,
            max_physical_window_size,
            max_window_size,
            window_size,
        )

        actual_foreground_color = raw_ui.GetForegroundColor(None, None)
        assert actual_foreground_color == foreground_color
        raw_ui.SetForegroundColor(None, None, Color.MAGENTA)
        assert isinstance(raw_ui.foreground_color, Color)
        assert raw_ui.foreground_color.value == Color.MAGENTA

        actual_background_color = raw_ui.GetBackgroundColor(None, None)
        assert actual_background_color == background_color
        raw_ui.SetBackgroundColor(None, None, Color.DARK_MAGENTA)
        assert isinstance(raw_ui.background_color, Color)
        assert raw_ui.background_color.value == Color.DARK_MAGENTA

        coordinates = GenericComplexObject()
        coordinates.extended_properties["x"] = 11
        coordinates.extended_properties["y"] = 12

        actual_cursor_position = raw_ui.GetCursorPosition(None, None)
        assert actual_cursor_position == cursor_position
        raw_ui.SetCursorPosition(None, None, coordinates)
        assert isinstance(raw_ui.cursor_position, Coordinates)
        assert raw_ui.cursor_position.x == 11
        assert raw_ui.cursor_position.y == 12

        actual_window_position = raw_ui.GetWindowPosition(None, None)
        assert actual_window_position == window_position
        raw_ui.SetWindowPosition(None, None, coordinates)
        assert isinstance(raw_ui.window_position, Coordinates)
        assert raw_ui.window_position.x == 11
        assert raw_ui.window_position.y == 12

        actual_cursor_size = raw_ui.GetCursorSize(None, None)
        assert actual_cursor_size == cursor_size
        raw_ui.SetCursorSize(None, None, 25)
        assert raw_ui.cursor_size == 25

        size = GenericComplexObject()
        size.extended_properties["height"] = 160
        size.extended_properties["width"] = 160

        actual_buffer_size = raw_ui.GetBufferSize(None, None)
        assert actual_buffer_size == buffer_size
        raw_ui.SetBufferSize(None, None, size)
        assert isinstance(raw_ui.buffer_size, Size)
        assert raw_ui.buffer_size.height == 160
        assert raw_ui.buffer_size.width == 160

        actual_window_size = raw_ui.GetWindowSize(None, None)
        assert actual_window_size == window_size
        raw_ui.SetWindowSize(None, None, size)
        assert isinstance(raw_ui.window_size, Size)
        assert raw_ui.window_size.height == 160
        assert raw_ui.window_size.width == 160

        actual_window_title = raw_ui.GetWindowTitle(None, None)
        assert actual_window_title == window_title
        raw_ui.SetWindowTitle(None, None, "new title")
        assert raw_ui.window_title == "new title"

        actual_max_window_size = raw_ui.GetMaxWindowSize(None, None)
        assert actual_max_window_size == max_window_size

        actual_physical_window_size = raw_ui.GetMaxPhysicalWindowSize(None, None)
        assert actual_physical_window_size == max_physical_window_size

        raw_ui.key_available = True
        actual_key_available = raw_ui.GetKeyAvailable(None, None)
        assert actual_key_available is True

        raw_ui.FlushInputBuffer(None, None)

        rectangle = GenericComplexObject()
        rectangle.extended_properties["left"] = 1
        rectangle.extended_properties["top"] = 2
        rectangle.extended_properties["right"] = 3
        rectangle.extended_properties["bottom"] = 4

        fill = GenericComplexObject()
        fill.extended_properties["character"] = "A"
        fill.extended_properties["foregroundColor"] = 12
        fill.extended_properties["backgroundColor"] = 10
        fill.extended_properties["bufferCellType"] = 0

        contents = GenericComplexObject()
        contents.extended_properties["mal"] = [2, 2]
        contents.extended_properties["mae"] = [[fill, fill], [fill, fill]]

        raw_ui.SetBufferContents1(None, None, rectangle, fill)
        raw_ui.SetBufferContents2(None, None, coordinates, contents)
        raw_ui.ScrollBufferContents(None, None, rectangle, coordinates, rectangle, fill)

    def test_ps_host_raw_ui_not_implemented(self):
        raw_ui = PSHostRawUserInterface(None, None, None, None, None, None, None, None, None, None)

        with pytest.raises(NotImplementedError):
            raw_ui.ReadKey(None, None)

        with pytest.raises(NotImplementedError):
            raw_ui.GetBufferContents(None, None, None)



import os
import time
import typing

from pypsrp.client import Client


def test_winrm() -> typing.Tuple[str, str, int]:
    server = os.environ["PYPSRP_SERVER"]
    username = os.environ["PYPSRP_USERNAME"]
    password = os.environ["PYPSRP_PASSWORD"]

    with Client(server, username=username, password=password, cert_validation=False) as c:
        return c.execute_cmd("whoami.exe")


def main() -> None:
    attempt = 1
    total_attempts = 5

    while True:
        print(f"Starting WinRM attempt {attempt}")

        try:
            stdout, stderr, rc = test_winrm()

        except Exception as e:
            print(f"Connection attempt {attempt} failed: {e!s}")

            if attempt == total_attempts:
                raise Exception(f"Exhausted {total_attempts} checking WinRM connection")

            print("Sleeping for 5 seconds before next attempt")
            attempt += 1
            time.sleep(5)

        else:
            print(f"Connection successful\nSTDOUT:\n{stdout}\nSTDERR:\n{stderr}\nRC: {rc}")
            break


if __name__ == "__main__":
    main()



from xmldiff import main as xml_diff


def assert_xml_diff(actual, expected, msg=None):
    # Only use xmldiff if it has been imported and both the xml messages
    # contain data
    if actual and expected:
        diff = xml_diff.diff_texts(actual, expected)
        assert len(diff) == 0, msg
    else:
        assert actual == expected, msg



import hashlib
import os
import tempfile
from collections import OrderedDict

import pytest

from pypsrp._utils import to_unicode
from pypsrp.client import Client
from pypsrp.exceptions import WinRMError
from pypsrp.powershell import PSDataStreams
from pypsrp.wsman import WSMan


class TestClient(object):
    def _get_client(self, wsman):
        # the connection object was already created as part of test fixture
        # we need to apply it to the Client object and set the values so the
        # test will work with existing responses
        client = Client("")
        client.wsman = wsman
        return client

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_copy_file"]], indirect=True)
    def test_client_copy_file(self, wsman_conn):
        client = self._get_client(wsman_conn)
        test_string = b"abcdefghijklmnopqrstuvwxyz"

        temp_file, path = tempfile.mkstemp()
        try:
            os.write(temp_file, test_string)
            actual = client.copy(path, "test_file")

            # run it a 2nd time to ensure it doesn't fail
            actual = client.copy(path, actual)
        finally:
            os.close(temp_file)
            os.remove(path)

        try:
            # verify the returned object is the full path
            assert actual.startswith("C:\\Users\\")

            actual_content = client.execute_cmd("powershell.exe Get-Content %s" % actual)[0].strip()
            assert actual_content == to_unicode(test_string)
        finally:
            client.execute_cmd("powershell Remove-Item -Path '%s'" % actual)

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_copy_file_empty"]], indirect=True)
    def test_client_copy_file_empty(self, wsman_conn):
        client = self._get_client(wsman_conn)

        temp_file, path = tempfile.mkstemp()
        try:
            actual = client.copy(path, "test_file")
        finally:
            os.close(temp_file)
            os.remove(path)

        try:
            # verify the returned object is the full path
            assert actual.startswith("C:\\Users\\")

            actual_content = client.execute_cmd("powershell.exe Get-Content %s" % actual)[0].strip()
            assert actual_content == ""
        finally:
            client.execute_cmd("powershell Remove-Item -Path '%s'" % actual)

    @pytest.mark.parametrize(
        "wsman_conn",
        # checks as to whether the correct number of calls
        # were sent and the remote requirements are too
        # variable to trust reliable
        [[False, "test_client_copy_file_really_large"]],
        indirect=True,
    )
    def test_client_copy_file_really_large(self, wsman_conn, monkeypatch, mocker):
        # in a mocked context the calculated size differs on a few variables
        # we will mock out that call and return the ones used in our existing
        # responses
        mock_calc = mocker.MagicMock()
        mock_calc.return_value = 113955
        monkeypatch.setattr(WSMan, "_calc_envelope_size", mock_calc)

        client = self._get_client(wsman_conn)

        test_string = b"abcdefghijklmnopqrstuvwxyz" * 20000

        temp_file, path = tempfile.mkstemp()
        try:
            os.write(temp_file, test_string)
            actual = client.copy(path, "test_file")
        finally:
            os.close(temp_file)
            os.remove(path)

        # verify the returned object is the full path
        assert actual == "C:\\Users\\vagrant\\Documents\\test_file"

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_copy_file_failure"]], indirect=True)
    def test_client_copy_file_failure(self, wsman_conn, monkeypatch, mocker):
        # set to a hash that is not the actual to verify the script will
        # fail in a hash mismatch scenario
        mock_hash = mocker.MagicMock()
        mock_hash.return_value.hexdigest.return_value = "c3499c2729730a7f807efb8676a92dcb6f8a3f8f"
        monkeypatch.setattr(hashlib, "sha1", mock_hash)

        client = self._get_client(wsman_conn)
        test_string = b"abcdefghijklmnopqrstuvwxyz"

        temp_file, path = tempfile.mkstemp()
        try:
            os.write(temp_file, test_string)
            with pytest.raises(WinRMError) as err:
                actual = client.copy(path, "test_file")
            expected_err = (
                "Failed to copy file: Transport failure, hash mismatch\r\n"
                "Actual: 32d10c7b8cf96570ca04ce37f2a19d84240d3a89\r\n"
                "Expected: c3499c2729730a7f807efb8676a92dcb6f8a3f8f"
            )
            assert expected_err in str(err.value)
        finally:
            os.close(temp_file)
            os.remove(path)

    @pytest.mark.parametrize(
        "wsman_conn",
        # Requires a specially crafted response to be returned
        [[False, "test_client_copy_file_warning"]],
        indirect=True,
    )
    def test_client_copy_file_warning(self, wsman_conn):
        client = self._get_client(wsman_conn)
        test_string = b"abcdefghijklmnopqrstuvwxyz"

        temp_file, path = tempfile.mkstemp()
        try:
            os.write(temp_file, test_string)
            with pytest.warns(
                Warning,
                match="Failed to disable MaximumAllowedMemory input size: You cannot call a "
                "method on a null-valued expression",
            ):
                actual = client.copy(path, "test_file")
            client.execute_cmd("powershell Remove-Item -Path '%s'" % actual)
        finally:
            os.close(temp_file)
            os.remove(path)

    @pytest.mark.parametrize(
        "wsman_conn",
        # Tested against a specific user name
        [[False, "test_client_copy_expand_vars"]],
        indirect=True,
    )
    def test_client_copy_expand_vars(self, wsman_conn):
        client = self._get_client(wsman_conn)
        test_string = b"abcdefghijklmnopqrstuvwxyz"

        temp_file, path = tempfile.mkstemp()
        try:
            os.write(temp_file, test_string)
            actual = client.copy(path, "%TEMP%\\test_file", expand_variables=True)

            # run it a 2nd time to ensure it doesn't fail
            actual = client.copy(path, actual)
        finally:
            os.close(temp_file)
            os.remove(path)

        try:
            # verify the returned object is the full path
            assert actual == "C:\\Users\\vagrant\\AppData\\Local\\Temp\\test_file"

            actual_content = client.execute_cmd("powershell.exe Get-Content %s" % actual)[0].strip()
            assert actual_content == to_unicode(test_string)
        finally:
            client.execute_cmd("powershell Remove-Item -Path '%s'" % actual)

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_execute_cmd"]], indirect=True)
    def test_client_execute_cmd(self, wsman_conn):
        client = self._get_client(wsman_conn)
        actual = client.execute_cmd("dir")
        actual_args = client.execute_cmd("echo abc")

        assert "Volume in drive C" in actual[0]
        assert actual[1] == ""
        assert actual[2] == 0

        assert actual_args[0] == "abc\r\n"
        assert actual_args[1] == ""
        assert actual_args[2] == 0

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_execute_cmd_environment"]], indirect=True)
    def test_client_execute_cmd_environment(self, wsman_conn):
        client = self._get_client(wsman_conn)

        env = OrderedDict(
            [
                ("string", "string value"),
                ("int", 1234),
                ("bool", True),
                ("double_quote", 'double " quote'),
                ("single_quote", "single ' quote"),
                ("hyphen - var", "abc @ 123"),
            ]
        )
        actual = client.execute_cmd("set", environment=env)
        actual_environment = actual[0].splitlines()
        for env_key, env_value in env.items():
            actual_env = next((e for e in actual_environment if e.startswith(env_key)), None)
            assert actual_env is not None
            assert actual_env == "%s=%s" % (env_key, env_value)

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_execute_ps"]], indirect=True)
    def test_client_execute_ps(self, wsman_conn):
        client = self._get_client(wsman_conn)

        expected_stdout = (
            "winrm\nRunning\n\nStatus   Name               "
            "DisplayName                           \n------"
            "   ----               -----------             "
            "              \nRunning  winrm              Wi"
            "ndows Remote Management (WS-Manag...\n\n"
        )
        actual = client.execute_ps("$serv = Get-Service -Name winrm; $serv.Name; $serv.Status; $serv")
        assert actual[0] == expected_stdout
        assert isinstance(actual[1], PSDataStreams)
        assert actual[2] is False

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_execute_ps_environment"]], indirect=True)
    def test_client_execute_ps_environment(self, wsman_conn):
        client = self._get_client(wsman_conn)

        env = OrderedDict(
            [
                ("string", "string value"),
                ("int", 1234),
                ("bool", True),
                ("double_quote", 'double " quote'),
                ("single_quote", "single ' quote"),
                ("hyphen - var", "abc @ 123"),
                ('_-(){}[]<>*+-/\\?"' "!@#$%^&|;:i,.`~0", '_-(){}[]<>*+-/\\?"' "!@#$%^&|;:i,.`~0"),
            ]
        )
        actual = client.execute_ps("cmd.exe /c set", environment=env)
        actual_environment = actual[0].splitlines()
        for env_key, env_value in env.items():
            actual_env = next((e for e in actual_environment if e.startswith(env_key)), None)
            assert actual_env is not None
            assert actual_env == "%s=%s" % (env_key, env_value)

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_execute_ps_failure"]], indirect=True)
    def test_client_execute_ps_failure(self, wsman_conn):
        client = self._get_client(wsman_conn)

        actual = client.execute_ps("Get-ServiceTypo -Name winrm")

        assert actual[0] == ""
        assert len(actual[1].error) == 1
        assert (
            str(actual[1].error[0]) == "The term 'Get-ServiceTypo' is not recognized as the name of a "
            "cmdlet, function, script file, or operable program. Check the "
            "spelling of the name, or if a path was included, verify that "
            "the path is correct and try again."
        )
        assert actual[2] is True

    @pytest.mark.parametrize(
        "wsman_conn",
        # means we don't need to create files on the
        # remote side
        [[False, "test_client_fetch_file"]],
        indirect=True,
    )
    def test_client_fetch_file(self, wsman_conn):
        client = self._get_client(wsman_conn)

        # file was created with
        # Set-Content -Path C:\temp\file.txt -Value ("abc`r`n" * 50000)

        temp_file, path = tempfile.mkstemp()
        os.close(temp_file)
        os.remove(path)
        try:
            client.fetch("C:\\temp\\file.txt", path)
            expected_hash = b"\x70\xe3\xbe\xa8\xcd\xb0\xd0\xc8\x83\xbc\xcf\xf5\x22\x89\x33\xd9\x33\xb8\x8a\x80"
            hash = hashlib.sha1()
            with open(path, "rb") as temp_file:
                while True:
                    data = temp_file.read(65536)
                    if not data:
                        break
                    hash.update(data)
            actual_hash = hash.digest()
            assert actual_hash == expected_hash
        finally:
            if os.path.exists(path):
                os.remove(path)

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_fetch_file_fail_dir"]], indirect=True)
    def test_client_fetch_file_fail_dir(self, wsman_conn):
        client = self._get_client(wsman_conn)
        with pytest.raises(WinRMError) as err:
            client.fetch("C:\\Windows", "")
        assert (
            str(err.value) == "Failed to fetch file C:\\Windows: The path at 'C:\\Windows' is "
            "a directory, src must be a file"
        )

    @pytest.mark.parametrize("wsman_conn", [[True, "test_client_fetch_file_fail_missing"]], indirect=True)
    def test_client_fetch_file_fail_missing(self, wsman_conn):
        client = self._get_client(wsman_conn)
        with pytest.raises(WinRMError) as err:
            client.fetch("C:\\fakefile.txt", "")
        assert str(err.value) == "Failed to fetch file C:\\fakefile.txt: The path at 'C:\\fakefile.txt' does not exist"

    @pytest.mark.parametrize(
        "wsman_conn",
        # use existing responses so I don't need to create
        # the file
        [[False, "test_client_fetch_file_hash_mismatch"]],
        indirect=True,
    )
    def test_client_fetch_file_hash_mismatch(self, wsman_conn, monkeypatch, mocker):
        # set to a hash that is not the actual to verify the script will
        # fail in a hash mismatch scenario
        mock_hash = mocker.MagicMock()
        mock_hash.return_value.hexdigest.return_value = "c3499c2729730a7f807efb8676a92dcb6f8a3f8f"
        monkeypatch.setattr(hashlib, "sha1", mock_hash)

        # file was created with
        # Set-Content -Path C:\temp\file.txt -Value ("abc`r`n" * 5)
        client = self._get_client(wsman_conn)
        with pytest.raises(WinRMError) as err:
            client.fetch("C:\\temp\\file.txt", "")
        assert (
            str(err.value) == "Failed to fetch file C:\\temp\\file.txt, hash mismatch\n"
            "Source: eec729d9a0fa275513bc44a4cb8d4ee973b81e1a\n"
            "Fetched: c3499c2729730a7f807efb8676a92dcb6f8a3f8f"
        )

    @pytest.mark.parametrize(
        "wsman_conn",
        # means we don't need to create files on the
        # remote side
        [[False, "test_client_fetch_file_expand_vars"]],
        indirect=True,
    )
    def test_client_fetch_file_expand_vars(self, wsman_conn):
        client = self._get_client(wsman_conn)

        # file was created with
        # Set-Content -Path "$env:TEMP\file.txt" -Value ("abc`r`n")

        temp_file, path = tempfile.mkstemp()
        os.close(temp_file)
        os.remove(path)
        try:
            client.fetch("%TEMP%\\file.txt", path, expand_variables=True)
            expected_hash = b"\x22\xc1\xa7\x8e\xc5\xad\xa1\xcd\x2f\x36\x65\xb5\x8b\x30\x49\x9e\x51\xa3\xb0\x29"

            hash = hashlib.sha1()
            with open(path, "rb") as temp_file:
                while True:
                    data = temp_file.read(65536)
                    if not data:
                        break
                    hash.update(data)
            actual_hash = hash.digest()
            assert actual_hash == expected_hash
        finally:
            if os.path.exists(path):
                os.remove(path)

    def test_sanitise_clixml_with_error(self):
        clixml_path = os.path.join(os.path.dirname(__file__), "data", "test_sanitise_clixml_with_error.xml")
        with open(clixml_path, "r") as fd:
            clixml = fd.read()

        expected = (
            "fake : The term 'fake' is not recognized as the name of a cmdlet, function, script file, or operable program. Check \r\n"
            "the spelling of the name, or if a path was included, verify that the path is correct and try again.\r\n"
            "At line:1 char:1\r\n"
            "+ fake cmdlet\r\n"
            "+ ~~~~\r\n"
            "    + CategoryInfo          : ObjectNotFound: (fake:String) [], CommandNotFoundException\r\n"
            "    + FullyQualifiedErrorId : CommandNotFoundException\r\n"
            " \r\n"
        )

        actual = Client.sanitise_clixml(clixml)
        assert actual == expected

    def test_sanitise_clixml_with_no_errors(self):
        clixml_path = os.path.join(os.path.dirname(__file__), "data", "test_sanitise_clixml_with_no_errors.xml")
        with open(clixml_path, "r") as fd:
            clixml = fd.read()

        expected = ""

        actual = Client.sanitise_clixml(clixml)
        assert actual == expected

    def test_sanitise_clixml_not_clixml(self):
        clixml = "stderr line"
        expected = clixml

        actual = Client.sanitise_clixml(clixml)
        assert actual == expected
